import java.util.*;

import de.cnc.domainparser.*;
import de.cnc.util.StringUtil;


/**
 * Codegenerator zum Erzeugen von Java-Klassen
 * fr hierarchische Datenstrukturen . <br/>
 * 
 * 
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 * 
 * @author Heiner Kcker
 * @version $Id$
 */
public class HierarchicDataGenerator
{

    /**
     * Arbeitsverzeichnis
     */
    static final String userDirStr = System.getProperty("user.dir");

    /**
     * Datei-Separator
     */
    static final String fileSeparatorStr = System.getProperty("file.separator");

    /**
     * Unterverzeichnis (und Package) fr die generierten Klassen
     */
    static final String outDirStr = "output";

    /**
     * Einrckung fr Codeformatierung . <br/>
     * Aktuell 4 Leerzeichen, es knnten auch eine
     * andere Anzahl Leerzeichen oder Tabs (\t) sein.
     */
    static final String INDENT = "    ";

    /**
     * Set mit allen definierten zu erzeugenden Klassen
     */
    static final HashSet struDefHashSet = new HashSet();

    /**
     * Start-Methode zum Erzeugen der Java-Klassen
     * fr hierarchische Datenstrukturen . <br/>
     * 
     * @param args Kommandozeilenparameter
     */
    public static void main(String[] args)
    {
        Unit defUnit = parseStructureDefinitionFile("Haus.strudef");

        checkStructureDefinition(defUnit);

        AbstractStruDef[] struDefArr = getStruDefArr(defUnit);

        generateClasses(struDefArr);
    }

    /**
     * Parsen der Strukturdefinitionsdatei mit dem DoaminParser.
     * 
     * @param pDefFileName Name der Strukturdefinitionsdatei
     * @return Unit (entspricht einer eingelesenen Datei)
     */
    private static Unit parseStructureDefinitionFile(final String pDefFileName)
    {
        final String defFileNameAndPath = userDirStr + fileSeparatorStr + pDefFileName;
        Unit defUnit = null;

        try
        {
            defUnit = Unit.parseFile(defFileNameAndPath, new DomainParserContext());
        } catch (DomainParseException e)
        {
            System.err.println(e.getMessage());
            e.printStackTrace();
            throw new RuntimeException(e);
        }
        // System.out.println( defUnit );

        defUnit.stripComments(false);
        defUnit.stripEmptySegments();
        // System.out.println( defUnit );
        return defUnit;
    }

    /**
     * Prfen der Struktuirdefinition.
     * 
     * @param pDefUnit eingeparste Unit (entspricht einer eingelesenen Datei)
     */
    private static void checkStructureDefinition(Unit pDefUnit)
    {
        final Segment mainSegment = pDefUnit.getSegment(0);

        processRecursiv(mainSegment, new CheckElementProcessor());
    }

    /**
     * Erzeugen eines Array's mit Strukturdefinitions-Objekten
     * aus der geparsten Strukturdefinitionsdatei.
     * 
     * @param pDefUnit eingeparste Unit (entspricht einer eingelesenen Datei)
     * @return Array mit den Strukturdefinitions-Objekten (Wurzel-Objekte)
     */
    private static AbstractStruDef[] getStruDefArr(Unit pDefUnit)
    {
        AbstractStruDef[] struDefArr = new AbstractStruDef[pDefUnit.segmentCount()];

        for (int i = 0; i < struDefArr.length; i++)
        {
            final Segment segment = pDefUnit.getSegment(i);

            struDefArr[i] = getStruDef(segment);
            // System.out.println( "struDefArr[i]\n" + struDefArr[i] );
        }
        struDefArr[0].isMainClass = true;
        return struDefArr;
    }

    /**
     * Rekursives Erzeugen eines hierarchischen Strukturdefinitions-Objekt's
     * aus der geparsten Strukturdefinitionsdatei.
     * 
     * @param pSegment eine Quell-Srukturdefinition
     * @return hierarchischen Strukturdefinitions-Objekt
     */
    private static AbstractStruDef getStruDef(final Segment pSegment)
    {
        int elemPos = 0;
        String javadocStr = null;

        if (pSegment.getElement(0) instanceof BracketComment)
        {
            javadocStr = ((BracketComment) pSegment.getElement(0)).getJavadocContent();
            elemPos++;
        }
        
        final Identifier classIdentifier = (Identifier) pSegment.getElement(elemPos);
        final String className = classIdentifier.getContent();

        elemPos++;

        final Identifier varIdentifier;
        final String varName;
        final boolean isListe;

        if (pSegment.getElement(elemPos) instanceof BracketArea) // BracketArea eckige Klammern
        {
            // Liste
            isListe = true;
            varIdentifier = (Identifier) pSegment.getElement(elemPos + 1);
            varName = varIdentifier.getContent();
        }
        else
        {
            // keine Liste
            isListe = false;
            varIdentifier = (Identifier) pSegment.getElement(elemPos);
            varName = varIdentifier.getContent();
        }
        boolean isComplex = pSegment.isBracesSegment();

        // create Root-Object 
        final AbstractStruDef retStruDef;

        if (!isListe && !isComplex)
        {
            retStruDef = new SimpleNoListStruDef();
        }
        else if (isListe && !isComplex)
        {
            retStruDef = new SimpleListStruDef();
        }
        else if (!isListe && isComplex)
        {
            retStruDef = new ComplexNoListStruDef();
        }
        else // if ( isListe && isComplex )
        {
            retStruDef = new ComplexListStruDef();
        }
        retStruDef.javadoc = javadocStr;
        retStruDef.className = className;
        retStruDef.varName = varName;

        if (isComplex)
        {

            // Prfen auf Einmalifkeit der zu erzeugenden Klasse
            if (struDefHashSet.contains(className))
            {
                try
                {
                    throw new DomainParseException("class " + className + " already defined", classIdentifier.getSourcePos());
                } catch (DomainParseException e)
                {
                    throw new RuntimeException(e);
                }
            }
            struDefHashSet.add(className);

            final AbstractComplexStruDef complexStruDef = (AbstractComplexStruDef) retStruDef;
            final BracesArea bracesArea = (BracesArea) pSegment.getElement(pSegment.elementCount() - 1);

            for (int i = 0; i < bracesArea.elementCount(); i++)
            {
                AbstractElement element = bracesArea.getSegment(i);
                // System.out.println( "generateOneClass: element.getClass(): " + element.getClass() );

                // if (element instanceof Segment)
                // {
                Segment subSegment = (Segment) element;

                complexStruDef.subStruDefList.add(getStruDef(subSegment));
                // }
            }
        }
        return retStruDef;
    }

    /**
     * Erzeugen der zu generierenden Java-Klassen
     * 
     * @param pStruDef hierarchisches Strukturdefinions-Objekt
     */
    private static void generateClasses(AbstractStruDef[] pStruDef)
    {
        for (int i = 0; i < pStruDef.length; i++)
        {
            generateOneClass(pStruDef[i]);
        }
    }

    /**
     * Erzeugen einer zu generierenden Java-Klasse.
     * 
     * @param pStruDef hierarchisches Strukturdefinions-Objekt
     */
    private static void generateOneClass(final AbstractStruDef pStruDef)
    {
        if (!pStruDef.isComplex())
        {
            // einfache Klassen mssen bereits existieren
            // es werden nur komplexe Klassen generiert
            return;
        }
        System.out.println("writeFile: " + pStruDef.className);
        final StringBuffer strBuff = new StringBuffer();

        strBuff.append("package output;\n\n");
        strBuff.append("import java.util.ArrayList;\n\n");
        // Javadoc
        if (pStruDef.javadoc != null)
        {
            strBuff.append("/**\n * " + pStruDef.javadoc + "\n */\n");
        }
        strBuff.append("public class " + pStruDef.className + "\n");
        strBuff.append("{\n\n");

        AbstractComplexStruDef complexStruDef = (AbstractComplexStruDef) pStruDef;
        ArrayList subStruDefList = complexStruDef.subStruDefList;

        for (int i = 0; i < subStruDefList.size(); i++)
        {
            final AbstractStruDef subStruDef = (AbstractStruDef) subStruDefList.get(i);

            if (subStruDef.isList())
            {
                // Liste

                // Javadoc
                if (subStruDef.javadoc != null)
                {
                    strBuff.append(INDENT + "/** list of " + subStruDef.javadoc + " */\n");
                }
                // Variablendeklaration
                strBuff.append(INDENT + "private ArrayList " + subStruDef.varName + " = new ArrayList();\n\n");

                final String capitalizedSubVarName = StringUtil.upper1(subStruDef.varName);

                // Javadoc
                if (subStruDef.javadoc != null)
                {
                    strBuff.append(INDENT + "/** add " + subStruDef.javadoc + " */\n");
                }
                // add-Methode
                strBuff.append(INDENT + "public void add" + capitalizedSubVarName + "(" + subStruDef.className + " new"
                        + capitalizedSubVarName + ")\n");
                strBuff.append(INDENT + "{\n");
                strBuff.append(INDENT + INDENT + subStruDef.varName + ".add(" + "new" + capitalizedSubVarName + ");\n");
                strBuff.append(INDENT + "}\n\n");

                // Javadoc
                if (subStruDef.javadoc != null)
                {
                    strBuff.append(INDENT + "/** get element count " + subStruDef.javadoc + " */\n");
                }
                // count-Methode
                strBuff.append(INDENT + "public int get" + capitalizedSubVarName + "Count()\n");
                strBuff.append(INDENT + "{\n");
                strBuff.append(INDENT + INDENT + "return " + subStruDef.varName + ".size();\n");
                strBuff.append(INDENT + "}\n\n");

                // Javadoc
                if (subStruDef.javadoc != null)
                {
                    strBuff.append(INDENT + "/** get element for specified index " + subStruDef.javadoc + " */\n");
                }
                // get-Methode
                strBuff.append(INDENT + "public " + subStruDef.className + " get" + capitalizedSubVarName + "(int pIndex)\n");
                strBuff.append(INDENT + "{\n");
                strBuff.append(INDENT + INDENT + "return (" + subStruDef.className + ") " + subStruDef.varName
                        + ".get(pIndex);\n");
                strBuff.append(INDENT + "}\n\n");
            }
            else
            {
                // keine Liste

                // Javadoc
                if (subStruDef.javadoc != null)
                {
                    strBuff.append(INDENT + "/** " + subStruDef.javadoc + " */\n");
                }
                // Variablendeklaration
                strBuff.append(INDENT + "private " + subStruDef.className + " " + subStruDef.varName + ";\n\n");

                final String capitalizedSubVarName = StringUtil.upper1(subStruDef.varName);

                // Javadoc
                if (subStruDef.javadoc != null)
                {
                    strBuff.append(INDENT + "/** set " + subStruDef.javadoc + " */\n");
                }
                // set-Methode
                strBuff.append(INDENT + "public void set" + capitalizedSubVarName + "(" + subStruDef.className + " new"
                        + capitalizedSubVarName + ")\n");
                strBuff.append(INDENT + "{\n");
                strBuff.append(INDENT + INDENT + subStruDef.varName + " = " + "new" + capitalizedSubVarName + ";\n");
                strBuff.append(INDENT + "}\n\n");

                // Javadoc
                if (subStruDef.javadoc != null)
                {
                    strBuff.append(INDENT + "/** get " + subStruDef.javadoc + " */\n");
                }
                // get-Methode
                strBuff.append(INDENT + "public " + subStruDef.className + " get" + capitalizedSubVarName + "()\n");
                strBuff.append(INDENT + "{\n");
                strBuff.append(INDENT + INDENT + "return " + subStruDef.varName + ";\n");
                strBuff.append(INDENT + "}\n\n");
            }

            // rekursiver Aufruf
            generateOneClass(subStruDef);
        }

        if (pStruDef.isMainClass)
        {
            strBuff.append(getCreateAndIterateArtefact(pStruDef));
        }

        strBuff.append("}");

        final String outputFileName = userDirStr + fileSeparatorStr + outDirStr + fileSeparatorStr + pStruDef.className
                + ".java";

        StringUtil.str2file(strBuff.toString(), outputFileName);

    }

    /**
     * Erzeugen der Artefakte zum Erzeugen und Durchlaufen
     * der definierten hierarchischen Datenstruktur.
     * 
     * @param pStruDef hierarchisches Strukturdefinions-Objekt
     * @return erzeugtes Artefakt
     */
    private static String getCreateAndIterateArtefact(final AbstractStruDef pStruDef)
    {
        StringBuffer strBuff = new StringBuffer();

        strBuff.append(INDENT + "public static void main(String[] args)\n");
        strBuff.append(INDENT + "{\n");

        // Erzeugen des Create-Artefaktes
        strBuff.append(getCreateArtefact(pStruDef, INDENT + INDENT));

        // Erzeugen des Iterate-Artefaktes
        strBuff.append(getIterateArtefact(pStruDef, INDENT + INDENT));

        strBuff.append(INDENT + "}\n\n");
        return strBuff.toString();
    }

    /**
     * Rekursives Erzeugen des Artefaktes zum Erzeugen
     * der definierten hierarchischen Datenstruktur.
     * 
     * @param pStruDef hierarchisches Strukturdefinions-Objekt
     * @param pIndent Einrckungs-String
     * @return erzeugtes Artefakt
     */
    private static String getCreateArtefact(// --
            final AbstractStruDef pStruDef, // --
            final String pIndent)
    {
        if (!pStruDef.isComplex())
        {
            // einfache Klassen mssen bereits existieren
            // es werden nur komplexe Klassen generiert
            return "";
        }
        final StringBuffer strBuff = new StringBuffer();

        if (pStruDef.isMainClass)
        {
            strBuff.append(pIndent + "//----------------------------------\n");
            strBuff.append(pIndent + "// Artefakt zum Aufbau der Struktur\n");
            strBuff.append(pIndent + "//----------------------------------\n");
            strBuff.append(pIndent + "boolean bFake = true; // bFake dient nur zum Bedienen der while-Schleifen\n\n");
        }

        strBuff.append(pIndent + pStruDef.className + " " + pStruDef.varName + " = new " + pStruDef.className + "();\n");
        if (pStruDef.isComplex())
        {
            //
            AbstractComplexStruDef complexStruDef = (AbstractComplexStruDef) pStruDef;
            ArrayList subStruDefList = complexStruDef.subStruDefList;

            for (int i = 0; i < subStruDefList.size(); i++)
            {
                final AbstractStruDef subStruDef = (AbstractStruDef) subStruDefList.get(i);

                if (subStruDef.isList())
                {
                    // Liste
                    strBuff.append(pIndent + "//Schleife " + subStruDef.className + " " + subStruDef.varName + "\n");
                    strBuff.append(pIndent + "while(bFake)\n");
                    strBuff.append(pIndent + "{\n");
                    // strBuff.append(pIndent + INDENT + subStruDef.className + " " + subStruDef.varName + " = new " + subStruDef.className + "();\n");
                    strBuff.append(getCreateArtefact(subStruDef, pIndent + INDENT));
                    if (subStruDef.isComplex())
                    {
                        strBuff.append(pIndent + INDENT + pStruDef.varName + ".add" + StringUtil.upper1(subStruDef.varName)
                                + "(" + subStruDef.varName + ");\n");
                    }
                    else
                    {
                        strBuff.append(pIndent + INDENT + pStruDef.varName + ".add" + StringUtil.upper1(subStruDef.varName)
                                + "(" + getFakeIniValue(subStruDef.className) + ");\n");
                    }

                    strBuff.append(pIndent + "}\n");
                }
                else
                {
                    // keine Liste
                    strBuff.append(pIndent + pStruDef.varName + ".set" + StringUtil.upper1(subStruDef.varName) + "("
                            + getFakeIniValue(subStruDef.className) + ");\n");
                }
            }
        }
        if (pStruDef.isMainClass)
        {
            strBuff.append("\n\n\n");
        }
        return strBuff.toString();
    }

    /**
     * Zurckgeben eines Standard-Initialisierungswertes
     * fr dias Artefakt zum Erzeugen der definierten
     * hierarchischen Datenstruktur.
     * 
     * @param pClassName Name derKlasse beziehungsweise des Types des 
     *     Initialisierungswertes
     * @return Initialisierungswert als String
     */
    private static String getFakeIniValue(String pClassName)
    {
        if ("int".equals(pClassName) || "long".equals(pClassName) || "short".equals(pClassName) || "byte".equals(pClassName)
                || "double".equals(pClassName) || "float".equals(pClassName))
        {
            return "0";
        }
        if ("char".equals(pClassName))
        {
            return "' '";
        }
        if ("boolean".equals(pClassName))
        {
            return "false";
        }

        if (struDefHashSet.contains(pClassName))
        {
            // eine in der Strukturdefinition enthaltene Klasse
            return "new " + pClassName + "()";
        }
        // eine Klasse und kein Elementar-Value
        // es wird mit null initialisiert
        return "(" + pClassName + ") null";
    }

    /**
     * Erzeugen des Artefaktes zum Durchlaufen
     * der definierten hierarchischen Datenstruktur.
     * 
     * @param pStruDef hierarchisches Strukturdefinions-Objekt
     * @param pIndent Einrckungs-String
     * @return erzeugtes Artefakt
     */
    private static String getIterateArtefact(// --
            final AbstractStruDef pStruDef, // --
            final String pIndent)
    {
        if (!pStruDef.isComplex())
        {
            // einfache Klassen mssen bereits existieren
            // es werden nur komplexe Klassen generiert
            return "";
        }
        final StringBuffer strBuff = new StringBuffer();

        if (pStruDef.isMainClass)
        {
            strBuff.append(pIndent + "//---------------------------------------\n");
            strBuff.append(pIndent + "// Artefakt zum Durchlaufen der Struktur\n");
            strBuff.append(pIndent + "//---------------------------------------\n");
        }

        if (pStruDef.isComplex())
        {
            //
            AbstractComplexStruDef complexStruDef = (AbstractComplexStruDef) pStruDef;
            ArrayList subStruDefList = complexStruDef.subStruDefList;

            for (int i = 0; i < subStruDefList.size(); i++)
            {
                final AbstractStruDef subStruDef = (AbstractStruDef) subStruDefList.get(i);

                if (subStruDef.isList())
                {
                    // Liste
                    strBuff.append(pIndent + "//Schleife " + subStruDef.className + " " + subStruDef.varName + "\n");
                    strBuff.append(pIndent + "for(int i" + StringUtil.upper1(subStruDef.varName) + " = 0 ; i"
                            + StringUtil.upper1(subStruDef.varName) + " < " + pStruDef.varName + ".get"
                            + StringUtil.upper1(subStruDef.varName) + "Count()" + " ; i"
                            + StringUtil.upper1(subStruDef.varName) + "++)\n");

                    strBuff.append(pIndent + "{\n");
                    if (subStruDef.isComplex())
                    {
                        strBuff.append(pIndent + INDENT + subStruDef.className + " " + subStruDef.varName + " = "
                                + pStruDef.varName + ".get" + StringUtil.upper1(subStruDef.varName) + "(i"
                                + StringUtil.upper1(subStruDef.varName) + ");\n");
                    }
                    else
                    {
                        strBuff.append(pIndent + INDENT + subStruDef.className + " " + subStruDef.varName + " = "
                                + pStruDef.varName + ".get" + StringUtil.upper1(subStruDef.varName) + "(i"
                                + StringUtil.upper1(subStruDef.varName) + ");\n");
                    }
                    // strBuff.append(pIndent + INDENT + subStruDef.className + " " + subStruDef.varName + " = new " + subStruDef.className + "();\n");
                    strBuff.append(getIterateArtefact(subStruDef, pIndent + INDENT));
                    strBuff.append(pIndent + "}\n");
                }
                else
                {// keine Liste
                    // strBuff.append(pIndent + pStruDef.varName + ".set" + StringUtil.upper1(subStruDef.varName) + "(" + getFakeIniValue( subStruDef.className ) + ");\n" );
                }
            }
        }
        return strBuff.toString();
    }

    /**
     * Rekursives Verarbeiten der Strukturdefinition fr
     * die hierarchische Datenstruktur . <br/>
     * 
     * @param pSegment ein DomainParserSegment
     * @param pElementProcessor Verarbeitungsklasse
     */
    private static void processRecursiv(// --
            final Segment pSegment, // --
            final AbstractElementProcessor pElementProcessor)
    {
        // final String className = pSegment.getElement(0).getContent();
        // System.out.println("generateOneClass: " + className);

        pElementProcessor.process(pSegment, false);

        // -----------------------------------------------------------------------
        // rekursives Abarbeiten der tiefer geschachtelten Strukturdefinitonen
        // -----------------------------------------------------------------------

        // Erfragen des letzten Elementes des bergebenen Segmentes
        BracesArea bracesArea = (BracesArea) pSegment.getElement(pSegment.elementCount() - 1);

        processRecursiv(bracesArea, pElementProcessor);
    }

    /**
     * Rekursives Verarbeiten der Strukturdefinition fr
     * die hierarchische Datenstruktur . <br/>
     * 
     * @param pBracesArea ein in geschwungene Klammern eingeschlossenes Stck 
     *     der Strukturdefinition
     * @param pElementProcessor Verarbeitungsklasse
     */
    private static void processRecursiv(
            final BracesArea pBracesArea,
            final AbstractElementProcessor pElementProcessor)
    {
        for (int i = 0; i < pBracesArea.elementCount(); i++)
        {
            AbstractElement element = pBracesArea.getSegment(i);

            // System.out.println( "generateOneClass: element.getClass(): " + element.getClass() );

            if (element instanceof Segment)
            {
                Segment segment = (Segment) element;

                if (segment.isBracesSegment())
                {
                    // Braces-Segmente (mit geschwungenen Klammern)
                    // sind nicht terminale Type, also separate Klassen
                    processRecursiv(segment, pElementProcessor);
                }
            }
        }
    }

    private static abstract class AbstractElementProcessor
    {

        /**
         * Verarbeitungs-Methode.
         * 
         * @param pSegment ein DomainParserSegment
         * @param pIsMainSegment ob das bergebene Segment ein Top_level (Root)- 
         *      Segment ist.
         */
        abstract void process(Segment pSegment, boolean pIsMainSegment);

    }


    /**
     * Verarbeitungklasse zum Prfen der Strukturdefinition.
     */
    private static class CheckElementProcessor extends AbstractElementProcessor
    {

        /**
         * Verarbeitungs-Methode.
         * 
         * @param pSegment ein DomainParserSegment
         * @param pIsMainSegment ob das bergebene Segment ein Top_level (Root)- 
         *      Segment ist.
         */
        void process(// --
                final Segment pSegment, // --
                final boolean pIsMainSegment)
        {
            try
            {
                if (pSegment.elementCount() < 1)
                {
                    // Bezeichner erwartet
                    throw new DomainParseException("Identifier or Javadoc expected", pSegment.getSourcePos());
                }

                if (!(pSegment.getElement(0) instanceof Identifier
                        || (pSegment.getElement(0) instanceof BracketComment
                                && ((BracketComment) pSegment.getElement(0)).isJavadoc())))
                {
                    // Bezeichner oder Javadoc erwartet
                    throw new DomainParseException("Identifier or Javadoc expected", pSegment.getElement(0).getSourcePos());
                }
                // final String className = pSegment.getElement(0).getContent();
                // System.out.println("check: " + className);

                if (pSegment.elementCount() < 2)
                {
                    // Bezeichner erwartet
                    throw new DomainParseException("Identifier or [] expected", pSegment.getEndSourcePos());
                }

                // @todo: hier noch weitere Prfungen einbauen

                // if ( mainSegment.elementCount() != 2 
                // || ! ( mainSegment.getElement( 0 ) instanceof Identifier ) 
                // || ! ( mainSegment.getElement( 1 ) instanceof BracesArea ) )
                // {
                // //Bezeichner und BracesArea erwartet
                // throw new DomainParseException( "Identifier and BracesArea expected" , mainSegment.getSourcePos() );
                // }
            } catch (DomainParseException e)
            {
                System.err.println(e.getMessage());
                e.printStackTrace();
                throw new RuntimeException(e);
            }
        }
    }


    /**
     * abstrakte Strukturdefinition
     */
    private static abstract class AbstractStruDef
    {

        /**
         * Zurckgeben, ob diese Strukturdefinition eine
         * Liste ist.
         * 
         * @return ob die bergebene Strukturdefinition eine Liste ist.
         */
        abstract boolean isList();

        /**
         * Zurckgeben, ob diese Strukturdefinition
         * untergeordnete Strukturdefinitionen
         * besitzt.
         * 
         * @return ob diese Strukturdefinition untergeordnete Strukturdefinitionen 
         *     besitzt.
         */
        abstract boolean isComplex();

        /**
         * die Javadoc dieser Strukturdefinition
         */
        String javadoc;

        /**
         * der Klassen-Name dieser Strukturdefinition
         */
        String className;

        /**
         * der variablen-Name dieser Strukturdefinition
         */
        String varName;

        /**
         * ob diese Strukturdefinition eine Wurzeldefinition ist
         */
        boolean isMainClass = false;

        /**
         * Debug
         * 
         * @return Debug
         */
        public String toString()
        {
            return // "\n" // + getClass().getName() //--
                    "[" // --
                    + "javadoc: " + javadoc // --
                    + ", className: " + className // --
                    + ", varName: " + varName // --
                    + ", isList: " + isList() // --
                    // + "]" 
                    ;
        }
    }


    /**
     * abstracte Strukturdefinition
     * ohne untergeordnete Strukturdefinitionen
     */
    private static abstract class AbstractSimpleStruDef extends AbstractStruDef
    {

        /**
         * Zurckgeben, ob diese Strukturdefinition
         * untergeordnete  Strukturdefinitionen
         * besitzt
         * 
         * @return ob diese Strukturdefinition untergeordnete  Strukturdefinitionen 
         *     besitzt
         */
        final boolean isComplex()
        {
            return false;
        }

        /**
         * Debug
         * 
         * @return Debug
         */
        public final String toString()
        {
            return super.toString() // --
                    + "]\n";
        }
    }


    /**
     * abstracte Strukturdefinition
     * mit untergeordneten Strukturdefinitionen
     */
    private static abstract class AbstractComplexStruDef extends AbstractStruDef
    {

        /**
         * Zurckgeben, ob diese Strukturdefinition
         * untergeordnete  Strukturdefinitionen
         * besitzt
         * 
         * @return ob diese Strukturdefinition untergeordnete  Strukturdefinitionen 
         *     besitzt
         */
        final boolean isComplex()
        {
            return true;
        }

        /**
         * untergeordnete  Strukturdefinitionen
         */
        final ArrayList subStruDefList = new ArrayList();

        /**
         * Debug
         * 
         * @return Debug
         */
        public final String toString()
        {
            return super.toString() // --
                    + ", subClassList: \n" + subStruDefList // --
                    + "]";
        }
    }


    /**
     * Strukturdefinition
     * ohne untergeordnete Strukturdefinitionen
     * keine Liste
     */
    private static class SimpleNoListStruDef extends AbstractSimpleStruDef
    {

        /**
         * Zurckgeben, ob diese Strukturdefinition
         * eine Liste ist.
         * 
         * @return ob diese Strukturdefinition eine Liste ist.
         */
        final boolean isList()
        {
            return false;
        }
    }


    /**
     * Strukturdefinition
     * ohne untergeordnete Strukturdefinitionen
     * eine Liste
     */
    private static class SimpleListStruDef extends AbstractSimpleStruDef
    {

        /**
         * Zurckgeben, ob diese Strukturdefinition
         * eine Liste ist.
         * 
         * @return ob diese Strukturdefinition eine Liste ist.
         */
        final boolean isList()
        {
            return true;
        }
    }


    /**
     * Strukturdefinition
     * mit untergeordneten Strukturdefinitionen
     * keine Liste
     */
    private static class ComplexNoListStruDef extends AbstractComplexStruDef
    {

        /**
         * Zurckgeben, ob diese Strukturdefinition
         * eine Liste ist.
         * 
         * @return ob diese Strukturdefinition eine Liste ist.
         */
        boolean isList()
        {
            return false;
        }

    }


    /**
     * Strukturdefinition
     * mit untergeordneten Strukturdefinitionen
     * eine Liste
     */
    private static class ComplexListStruDef extends AbstractComplexStruDef
    {

        /**
         * Zurckgeben, ob diese Strukturdefinition
         * eine Liste ist.
         * 
         * @return ob diese Strukturdefinition eine Liste ist.
         */
        boolean isList()
        {
            return true;
        }
    }

}
